==> file4life-1.1.5-source.tgz: file4life-1.1.5/src/F4L/Actions/Backup.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Actions.Backup (
  autoBackup
  , backupPath
  , checkBackup
  , restore
  ) where

import F4L
import F4L.Actions.Common
import F4L.Actions.Files
import F4L.Actions.Ops
import F4L.Actions.Storages
import F4L.BufferMonad
import F4L.F4LMonad
import F4L.FSMonad hiding ( outputLineMsg, cleanLine )
import F4L.Knowledge
import F4L.Op
import F4L.SeenTree as ST hiding ( removeFile )
import F4L.Utils hiding( fileSize )

import Control.Lens
import Control.Monad
import Control.Monad.Error( ErrorT(..), throwError, noMsg )
import Control.Monad.Reader
import Control.Monad.State.Strict
import Data.List( sortBy )
import qualified Data.Map.Strict as M -- TODO: just use SM
import qualified Data.Map.Strict as SM
import qualified Data.Set as S
import System.FilePath( combine, takeDirectory, joinPath, splitDirectories )


-- | Backup at most maxSize bytes depending on priority, size and known copies
autoBackup :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k ) 
              => FilePath -> Int -> F4LMonad k m ()
autoBackup dest maxSize = do
  fullData <- use st_knowledge  
  -- Fids to backup by reverse weight
  let fids = map fst $ sortBy ( compare `byRev` benefitOverCost ) $ M.toList $ view kd_files fullData
  backup fids dest maxSize

  where
    -- TODO: find a more reasonable fun
    benefitOverCost :: ( FileId, FileData ) -> Float
    benefitOverCost ( fid, fd ) =
      ( fromIntegral $ pri $ fd ^. fd_priority )
      * p ^^ ( S.size $ fd ^. fd_seen ) * ( 1 - p )
      / fromIntegral ( max 1 $ fidDiskSpace fid )

    p = 0.2 -- Probability one copy of the file disappears

    pri :: Priority -> Int
    pri ( Priority i ) = 10 ^ i
    pri DefaultPriority = 1
    pri AsManyAsPossible = pri ( Priority 10 )
    pri _ = 0

    byRev :: ( a -> a -> c ) -> ( b -> a ) -> ( b -> b -> c )
    byRev f t = \ a b -> f ( t b ) ( t a )
  

-- | Backup the whole contents of sid:path into dest
backupPath :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k )
              => StorageId -> FilePath -> FilePath -> F4LMonad k m ()
backupPath sid path dest = do
  st <- use $ st_knowledge . kd_contents . to ( M.findWithDefault ST.empty sid )
  let fids = map snd $ subFiles ( splitDirectories path ) st
  backup fids dest 1000000000000


backup :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k )
          => [ FileId ] -> FilePath -> Int -> F4LMonad k m ()
backup fids dest maxSize = onErrorWarn $ do
  lift $ st_storages .= SM.empty
  fullData <- lift $ use st_knowledge
  envid <- lift $ use st_envid
  say Timid $ "Env: " ++ envid
  storages <- lift storagesForEnv
  createDirectoryIfMissing True dest
  sid <- forceGetStorageId dest
  let seenfids = S.fromList $ ST.fids $ M.findWithDefault ST.empty sid $ view kd_contents fullData
  say Chatty $ "Seen fids: " ++ show seenfids
  let fidDataMap = view kd_files fullData
  let fidsData = map ( \ fid -> ( fid, M.findWithDefault noFileData fid fidDataMap ) ) fids
  let newfids = flip filter fidsData $ not . ( `S.member` seenfids ) . fst
  lift $ do
    withStorage sid dest
    runStateT ( backupFids storages maxSize newfids ) (0, 0)
    cleanLine
  lift $ st_storages .= SM.empty
  where
    backupFids :: ( DelFSMonad m, BufferMonad m, Knowledge k ) => SM.Map StorageId [ FilePath ] -> Int -> [(FileId, FileData)] -> StateT (Int, Int) ( F4LMonad k m ) ()
    backupFids storages s fids = -- liftIO ( putStrLn $ "Backing up size: " ++ show s ) >>
      let nfids = dropWhile ( ( > s ) . fidDiskSpace . fst ) fids
      in
       case nfids of
         ( fid, dat ) : rest -> do
           let ( dp, destPath ) = destFidPath fid
           isThere <- doesFileExist destPath
           case isThere of
             True -> do -- File already present, but not recorded
               isOld <- runErrorT $ do
                 s <- fileSize destPath
                 when ( s /= fid_length fid ) $ throwError noMsg
                 fileFid <- fileId destPath
                 when ( fileFid /= fid ) $ throwError noMsg
               case isOld of
                 Left _ -> say Quiet ( "Skipping " ++ destPath ++ " present with non-matching FileId" ) >> backupFids storages s rest
                 Right _ -> do
                   say Timid ( "Found " ++ destPath ++ " already present; adding" ) >> backupFids storages s rest
                   lift $ do
                     setFid fid
                     addOp BackupPath
                     addOp Seen
             False -> do
               lift $ onErrorWarn $ createDirectoryIfMissing True dp
               ( nfiles, donesize ) <- get
               let candidates = fileCandidates storages dat
               res <- lift $ runErrorT $ checkedCopyFileCandidates fid candidates destPath $ hShow nfiles ++ " " ++ hShow donesize ++ "\t"
               -- TODO: deal with disk full error
               case res of
                 Right _ -> do
                   _1 += 1
                   _2 += fidDiskSpace fid
                   lift $ do
                     setFid fid
                     addOp BackupPath
                     addOp Seen
                   backupFids storages ( s - fidDiskSpace fid ) rest
                 Left _err -> do
                   -- say Quiet $ "Skipping on error: " ++ _err
                   backupFids storages s rest
         _ ->
           return ()

    destFidPath fid =
      let ( p, fn ) = fidToBDest fid
      in ( combine dest p, combine dest $ combine p fn )


-- | Fail on errors with DB or destination
--   Report result: file id, destination path, flag
--   flag = Nothing -> file was already present
--   flag = Just True -> file restored
--   flag = Just False -> file not restored
restore :: ( DelFSMonad m, BufferMonad m, KnownDataClass k, Knowledge k ) 
           => Bool -> StorageId -> FilePath -> FilePath -> ErrorT Error ( F4LMonad k m ) [(FileId, FilePath, Maybe Bool)]
restore recorded sid srcpath dest = do
  lift $ st_storages .= SM.empty
  createDirectoryIfMissing True dest
  -- Get destination storage data
  srp <- ( if recorded then forceGetStorage else fakeGetStorage ) dest
  lift $ do
    let destsid = rp_sid srp
    let sPath = rp_sPath srp
    let relDest = rp_relPath srp
    storages <- storagesForEnv
    let pfx = splitDirectories srcpath
    toCopy <- use $ st_knowledge . kd_contents . to ( M.findWithDefault ST.empty sid ) . to ( subFiles pfx )
    say Timid $ show $ toCopy
    when recorded $ withStorage destsid dest
    kdFiles <- use $ st_knowledge . kd_files
    res <- mapM ( \ ( path, fid ) -> tryCopy storages kdFiles fid sPath $ joinPath $ relDest : path ) toCopy
    cleanLine
    st_storages .= SM.empty    
    return res
  where
    tryCopy :: ( DelFSMonad m, BufferMonad m, Knowledge k )
               => SM.Map StorageId [FilePath] -> SM.Map FileId FileData -> FileId -> FilePath -> FilePath 
               -> F4LMonad k m (FileId, FilePath, Maybe Bool)
    tryCopy storages files fid storagePath relDestPath = do
      let destPath = joinPath [ storagePath, relDestPath ]
      say Timid $ "Copying " ++ show fid ++ " to " ++ destPath
      outputLineMsg $ "Copying " ++ show fid ++ " to " ++ destPath
      isThere <- doesFileExist destPath
      case isThere of
        True -> do
          say Chatty "Skipping"
          return ( fid, relDestPath, Nothing )
        False -> do
          res <- runErrorT $ do
            createDirectoryIfMissing True $ takeDirectory destPath
            let mdat = M.lookup fid files
            case mdat of
              Just dat -> do
                let candidates = fileCandidates storages dat
                copied <- lift $ runErrorT $ checkedCopyFileCandidates fid candidates destPath ""
                case copied of
                  Right _ -> do
                    when recorded $ lift $ do
                      setFid fid
                      setPath relDestPath
                      addOp Seen
                    return ( fid, relDestPath, Just True )
                  Left err -> do
                    say Chatty $ "Skipping: not found anywhere (" ++ err ++ ")"
                    return ( fid, relDestPath, Just False )
              Nothing -> do
                say Chatty "Skipping: not found anywhere"
                return ( fid, relDestPath, Just False )
          return $ either ( const ( fid, relDestPath, Just False ) ) id res


checkBackup :: FSMonad m => Bool -> FilePath -> m () -- [ ( FilePath, FileId, String ) ]
checkBackup quick dir = recur
  where
    showMsg ( fn, _fid, msg ) = outputMsg $ joinPath [ dir, fn ] ++ "\t" ++ msg
    recur = do
      l1 <- onErrorUseDefault [] $ getCleanDirectoryContents dir
      mapM_ recur1 $ filter ( (==) 2 . length ) l1
    recur1 l1 = do
      l2 <- onErrorUseDefault [] $ getCleanDirectoryContents $ joinPath [ dir, l1 ]
      mapM_ ( recur2 l1 ) $ filter ( (==) 2 . length ) l2
    recur2 l1 l2 = do
      fn <- onErrorUseDefault [] $ getCleanDirectoryContents $ joinPath [ dir, l1, l2 ]
      mapM_ check [ joinPath [ l1, l2, x ] | x <- fn ]
    check fn = do
      case fidFromBDest fn of
        Nothing ->
          return ()
        Just fid -> do
          let fullPath = joinPath [ dir, fn ]
          s <- onErrorUseDefault (-1) $ fileSize fullPath
          if s /= fid_length fid
            then showMsg ( fn, fid, "Length mismatch" )
            else unless quick $ do
            actualFid <- onErrorUseDefault Nothing $ liftM Just $ fileId fullPath
            case actualFid of
              Nothing -> showMsg ( fn, fid, "Unable to compute file ID" )
              Just x | x /= fid -> showMsg ( fn, fid, "File ID mismatch" )
              _ -> return ()

==> file4life-1.2.0-source.tgz: file4life-1.2.0/src/F4L/Actions/Backup.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Actions.Backup (
  autoBackup
  , backupPath
  , checkBackup
  , restore
  ) where

import F4L
import F4L.Actions.Common
import F4L.Actions.Files
import F4L.Actions.Ops
import F4L.Actions.Storages
import F4L.BufferMonad
import F4L.F4LMonad
import F4L.FidStore
import F4L.FileTree as FT
import F4L.FSMonad hiding ( outputLineMsg, cleanLine )
import F4L.Knowledge
import F4L.Op
import F4L.StoryTree as ST hiding ( removeFile )
import F4L.Utils hiding( fileSize )

import Lens.Micro.Platform
import Control.Monad
import Control.Monad.Trans.Except( ExceptT(..), runExceptT )
import Control.Monad.Reader
import Control.Monad.State.Strict
import qualified Data.IntMap.Strict as IM
import Data.List( sortBy )
import qualified Data.Map.Strict as M -- TODO: just use SM
import qualified Data.Map.Strict as SM
import qualified Data.Set as S
import System.FilePath( combine, takeDirectory, joinPath, splitDirectories )


-- | Backup at most maxSize bytes depending on priority, size and known copies
autoBackup :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k ) 
              => FilePath -> Int -> F4LMonad k m ()
autoBackup dest maxSize = do
  fullData <- use st_knowledge
  fidstore <- use $ st_opEnv . openv_fids
  -- Fids to backup by reverse weight
  let fids = map fst $ sortBy ( compare `byRev` benefitOverCost ) $ map ( mapFst $ fidByIndex' fidstore ) $ IM.toList $ view kd_files fullData
  backup fids dest maxSize

  where
    -- TODO: find a more reasonable fun
    benefitOverCost :: ( FileId, FileData ) -> Float
    benefitOverCost ( fid, fd ) =
      ( fromIntegral $ pri $ fd ^. fd_priority )
      * p ^^ ( S.size $ fd ^. fd_seen ) * ( 1 - p )
      / fromIntegral ( max 1 $ fidDiskSpace fid )

    p = 0.2 -- Probability one copy of the file disappears

    pri :: Priority -> Int
    pri ( Priority i ) = 10 ^ i
    pri DefaultPriority = 1
    pri AsManyAsPossible = pri ( Priority 10 )
    pri _ = 0

    byRev :: ( a -> a -> c ) -> ( b -> a ) -> ( b -> b -> c )
    byRev f t = \ a b -> f ( t b ) ( t a )
  

-- | Backup the whole contents of sid:path into dest
backupPath :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k )
              => StorageId -> FilePath -> FilePath -> F4LMonad k m ()
backupPath sid path dest = do
  st <- use $ st_knowledge . kd_contents . to ( M.findWithDefault ST.empty sid )
  fidstore <- use $ st_opEnv . openv_fids
  let fids = map ( fidByIndex' fidstore . snd ) $ subFiles ( splitDirectories path ) $ atTime endOfTime st
  backup fids dest 1000000000000


backup :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k )
          => [ FileId ] -> FilePath -> Int -> F4LMonad k m ()
backup fids dest maxSize = onErrorWarn $ do
  lift $ st_storages .= SM.empty
  fullData <- lift $ use st_knowledge
  envid <- lift $ use st_envid
  say Timid $ "Env: " ++ envid
  storages <- lift storagesForEnv
  createDirectoryIfMissing True dest
  sid <- forceGetStorageId dest
  let seenidxs = S.fromList $ FT.files $ atTime endOfTime $ M.findWithDefault ST.empty sid $ view kd_contents fullData
  say SayAll $ "Seen fid indices: " ++ show seenidxs
  let idxDataMap = view kd_files fullData
  fidstore <- use $ st_opEnv . openv_fids
  let idxData = map ( \ idx -> ( idx, IM.findWithDefault noFileData idx idxDataMap ) ) $ map ( fidIndex' fidstore ) fids
  let newfids = map ( mapFst $ fidByIndex' fidstore ) $ flip filter idxData $ not . ( `S.member` seenidxs ) . fst
  lift $ do
    withStorage sid dest
    runStateT ( backupFids storages maxSize newfids ) (0, 0)
    cleanLine
  lift $ st_storages .= SM.empty
  where
    backupFids :: ( DelFSMonad m, BufferMonad m, Knowledge k ) => SM.Map StorageId [ FilePath ] -> Int -> [(FileId, FileData)] -> StateT (Int, Int) ( F4LMonad k m ) ()
    backupFids storages s fids = -- liftIO ( putStrLn $ "Backing up size: " ++ show s ) >>
      let nfids = dropWhile ( ( > s ) . fidDiskSpace . fst ) fids
      in
       case nfids of
         ( fid, dat ) : rest -> do
           let ( dp, destPath ) = destFidPath fid
           isThere <- doesFileExist destPath
           if isThere then
             do -- File already present, but not recorded
               isOld <- runExceptT $ do
                 s <- fileSize destPath
                 when ( s /= fid_length fid ) noMsg
                 fileFid <- fileId destPath
                 when ( fileFid /= fid ) noMsg
               case isOld of
                 Left _ -> say Quiet ( "Skipping " ++ destPath ++ " present with non-matching FileId" ) >> backupFids storages s rest
                 Right _ -> do
                   say Timid ( "Found " ++ destPath ++ " already present; adding" ) >> backupFids storages s rest
                   lift $ do
                     setFid fid
                     addOp BackupPath
                     addOp Seen
             else
             do
               lift $ onErrorWarn $ createDirectoryIfMissing True dp
               ( nfiles, donesize ) <- get
               let candidates = fileCandidates storages dat
               res <- lift $ runExceptT $ checkedCopyFileCandidates fid candidates destPath $ hShow nfiles ++ " " ++ hShow donesize ++ "\t"
               -- TODO: deal with disk full error
               case res of
                 Right _ -> do
                   _1 += 1
                   _2 += fidDiskSpace fid
                   lift $ do
                     setFid fid
                     addOp BackupPath
                     addOp Seen
                   backupFids storages ( s - fidDiskSpace fid ) rest
                 Left _err -> do
                   say Chatty $ "Skipping on error: " ++ _err
                   backupFids storages s rest
         _ ->
           return ()

    destFidPath fid =
      let ( p, fn ) = fidToBDest fid
      in ( combine dest p, combine dest $ combine p fn )


-- | Fail on errors with DB or destination
--   Report result: file id, destination path, flag
--   flag = Nothing -> file was already present
--   flag = Just True -> file restored
--   flag = Just False -> file not restored
restore :: ( DelFSMonad m, BufferMonad m, KnownDataClass k, Knowledge k ) 
           => Bool -> StorageId -> FilePath -> FilePath -> ExceptT Error ( F4LMonad k m ) [(FileId, FilePath, Maybe Bool)]
restore recorded sid srcpath dest = do
  lift $ st_storages .= SM.empty
  createDirectoryIfMissing True dest
  -- Get destination storage data
  srp <- ( if recorded then forceGetStorage else fakeGetStorage ) dest
  lift $ do
    let destsid = rp_sid srp
    let sPath = rp_sPath srp
    let relDest = rp_relPath srp
    storages <- storagesForEnv
    let pfx = splitDirectories srcpath
    toCopy <- use $ st_knowledge . kd_contents . to ( M.findWithDefault ST.empty sid ) . to ( subFiles pfx . atTime endOfTime )
    say Timid $ show $ toCopy
    when recorded $ withStorage destsid dest
    kdFiles <- use $ st_knowledge . kd_files
    fidstore <- use $ st_opEnv . openv_fids
    res <- mapM ( \ ( path, idx ) -> tryCopy storages kdFiles idx ( fidByIndex' fidstore idx ) sPath $ joinPath $ relDest : path ) toCopy
    cleanLine
    st_storages .= SM.empty    
    return res
  where
    tryCopy :: ( DelFSMonad m, BufferMonad m, Knowledge k )
               => SM.Map StorageId [FilePath] -> IM.IntMap FileData -> FidIndex -> FileId -> FilePath -> FilePath 
               -> F4LMonad k m (FileId, FilePath, Maybe Bool)
    tryCopy storages files idx fid storagePath relDestPath = do
      let destPath = joinPath [ storagePath, relDestPath ]
      say Timid $ "Copying " ++ show fid ++ " to " ++ destPath
      outputLineMsg $ "Copying " ++ show fid ++ " to " ++ destPath
      isThere <- doesFileExist destPath
      if isThere then
        do
          say Chatty "Skipping"
          return ( fid, relDestPath, Nothing )
        else
        do
          res <- runExceptT $ do
            createDirectoryIfMissing True $ takeDirectory destPath
            let mdat = IM.lookup idx files
            case mdat of
              Just dat -> do
                let candidates = fileCandidates storages dat
                copied <- lift $ runExceptT $ checkedCopyFileCandidates fid candidates destPath ""
                case copied of
                  Right _ -> do
                    when recorded $ lift $ do
                      setFid fid
                      setPath relDestPath
                      addOp Seen
                    return ( fid, relDestPath, Just True )
                  Left err -> do
                    say Chatty $ "Skipping: not found anywhere (" ++ err ++ ")"
                    return ( fid, relDestPath, Just False )
              Nothing -> do
                say Chatty "Skipping: not found anywhere"
                return ( fid, relDestPath, Just False )
          return $ either ( const ( fid, relDestPath, Just False ) ) id res


checkBackup :: FSMonad m => Bool -> FilePath -> m () -- [ ( FilePath, FileId, String ) ]
checkBackup quick dir = recur
  where
    showMsg ( fn, _fid, msg ) = outputMsg $ joinPath [ dir, fn ] ++ "\t" ++ msg
    recur = do
      l1 <- onErrorUseDefault [] $ getCleanDirectoryContents dir
      mapM_ recur1 $ filter ( (==) 2 . length ) l1
    recur1 l1 = do
      l2 <- onErrorUseDefault [] $ getCleanDirectoryContents $ joinPath [ dir, l1 ]
      mapM_ ( recur2 l1 ) $ filter ( (==) 2 . length ) l2
    recur2 l1 l2 = do
      fn <- onErrorUseDefault [] $ getCleanDirectoryContents $ joinPath [ dir, l1, l2 ]
      mapM_ check [ joinPath [ l1, l2, x ] | x <- fn ]
    check fn = do
      case fidFromBDest fn of
        Nothing ->
          return ()
        Just fid -> do
          let fullPath = joinPath [ dir, fn ]
          s <- onErrorUseDefault (-1) $ fileSize fullPath
          if s /= fid_length fid
            then showMsg ( fn, fid, "Length mismatch" )
            else unless quick $ do
            actualFid <- onErrorUseDefault Nothing $ liftM Just $ fileId fullPath
            case actualFid of
              Nothing -> showMsg ( fn, fid, "Unable to compute file ID" )
              Just x | x /= fid -> showMsg ( fn, fid, "File ID mismatch" )
              _ -> return ()
